# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
import os
sys.path.append(os.path.join(os.path.dirname(__file__), '../src/diresa'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'DIRESA'
copyright = '2024, Geert De Paepe, ETRO VUB'
author = 'Geert De Paepe'


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc', 'sphinx_rtd_theme', 'sphinx_copybutton']
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['_static']
html_theme = "sphinx_rtd_theme"
html_theme_options = {"logo_only": True}
html_logo = "images/diresa.png"
html_show_sourcelink = True
gitlab_url = "https://gitlab.com/etrovub/ai4wcm/public/diresa"
autodoc_default_options = {
    'member-order': 'bysource',
    'undoc-members': False,
    'special-members': '__init__',
}
autodoc_mock_imports = ["tensorflow", "tensorflow_probability", "keras", "diresa"]




#!/usr/bin/env python3
"""
DIRESA callback classes/functions

:Author:  Geert De Paepe
:Email:   geert.de.paepe@vub.be
:License: MIT License
"""

from tensorflow.keras.backend import get_value, set_value
from tensorflow.keras.callbacks import Callback


class LossWeightAnnealing(Callback):
    """
    https://keras.io/guides/writing_your_own_callbacks/
    https://medium.com/dive-into-ml-ai/adaptive-weighing-of-loss-functions-for-multiple-output-keras-models-71a1b0aca66e
    """
    def __init__(self, weight, loss_name="val_Cov_loss", target_loss=0.000003, anneal_step=0.1, start_epoch=5):
        """
        :param weight: keras.backend.variable with initial loss weight
        :param loss_name: name of the loss function to apply the annealing
        :param target_loss: target loss, weight is increased until loss < target
        :param anneal_step: annealing step size for increasing weight
        :param start_epoch: epoch from which annealing starts
        """
        super().__init__()
        self.weight = weight
        self.loss_name = loss_name
        self.target_loss = target_loss
        self.anneal_step = anneal_step
        self.start_epoch = start_epoch
        self.stop_annealing = None

    def on_epoch_end(self, epoch, logs=None):
        """
        Executed during training on each epoch end
        Increases weight with anneal_step as from start_epoch until unweighted loss is smaller than target
        Prints loss weight, reconstruction loss and unweighted loss

        :param epoch: number of epoch
        :param logs: dict containing the loss values
        """
        weight = float(get_value(self.weight))
        if weight > 0.00001:
            unweighted_loss = logs[self.loss_name] / weight
        else:
            unweighted_loss = 1.
        if unweighted_loss < self.target_loss and self.stop_annealing is None:
            self.stop_annealing = epoch
        if (epoch + 2) >= self.start_epoch and self.stop_annealing is None:
            set_value(self.weight, weight + self.anneal_step)
        print("Covariance loss weight: {:g}, Validation Covariance loss (unweighted): {:g}."
              .format(weight, unweighted_loss))

    def on_epoch_begin(self, epoch, logs=None):
        """
        Divides learning rate by 2 every 10th epoch after annealing is stopped

        :param epoch: number of epoch
        :param logs: dict containing the loss values
        """
        if self.stop_annealing is not None:
            if epoch > self.stop_annealing + 10 and (epoch - self.stop_annealing) % 10 == 0:
                lr = float(get_value(self.model.optimizer.learning_rate))
                lr /= 2
                set_value(self.model.optimizer.lr, lr)
                print("\nEpoch %05d: Learning rate is %f" % (epoch, lr))



#!/usr/bin/env python3
"""
:Author:  Geert De Paepe
:Email:   geert.de.paepe@vub.be
:License: MIT License
"""

from math import isclose
import tensorflow as tf
from diresa.toolbox import covariance, r2_score
import numpy as np


def test_covariance():
    # test data
    data = tf.constant([[2.1, 2.5, 3.6], [3.4, 4.1, 5.2], [4.5, 5.6, 6.7]], dtype=tf.float32)
    # numpy: np.cov(data, bias=True, rowvar=False)
    cov = covariance(data)
    expected_cov = [[0.96222230, 1.24111112, 1.24111112],
                    [1.24111112, 1.60222212, 1.60222212],
                    [1.24111112, 1.60222212, 1.60222212]]
    assert isclose(tf.reduce_sum(cov), sum(map(sum, expected_cov)), abs_tol=1e-5), f"Expected {cov}, but got {expected_cov}"


def test_r2_score():
    # test data
    y_true = tf.constant([[3, -0.5, 2, 7], [2, 0, 0, 1]], dtype=tf.float32)
    y_pred = tf.constant([[2.5, 0.0, 2, 8], [2, 0.5, 0, 1]], dtype=tf.float32)
    # scikit-learn: r2_score(y_true, y_pred, multioutput='variance_weighted')
    r2 = r2_score(y_true, y_pred)
    expected_r2 = 0.9151515
    assert isclose(r2, expected_r2, abs_tol=1e-5), f"Expected {expected_r2}, but got {r2.numpy()}"


if __name__ == "__main__":
    test_r2_score()



#!/usr/bin/env python3

import numpy as np
from diresa.models import build_diresa
from diresa.loss import mse_dist_loss, LatentCovLoss
from diresa.layers import DistLayer
from tensorflow.keras import backend as K
from diresa.callback import LossWeightAnnealing
from diresa.toolbox import cut_sub_model
from tensorflow.keras.models import load_model
from tensorflow.keras.regularizers import L2


def load_dataset():
    data_file = "docs/lorenz.csv"
    data = np.loadtxt(data_file, delimiter=",")
    train = data[:5000]
    val = data[5000:6000]
    id_train = np.argsort((np.random.random(train.shape[0])))
    id_val = np.argsort((np.random.random(val.shape[0])))
    train_twin = train[id_train]
    val_twin = val[id_val]
    return train, train_twin, val, val_twin


def test_diresa():
    train, train_twin, val, val_twin = load_dataset()

    # build the model
    diresa = build_diresa(input_shape=(3,), dense_units=(40, 20, 2))
    cov_weight = K.variable(0.)
    diresa.compile(loss=['MSE', LatentCovLoss(cov_weight), mse_dist_loss], loss_weights=[1., 1., 1.5])

    # train the model
    diresa.fit((train, train_twin), (train, train, train),
               validation_data=((val, val_twin), (val, val, val)),
               epochs=5, batch_size=512, shuffle=True, verbose=2)

    # encoder submodel
    cut_sub_model(diresa, 'Encoder')

    # save and open model
    diresa.save('diresa.keras')
    load_model('diresa.keras')


def test_annealing_diresa():
    train, train_twin, val, val_twin = load_dataset()

    # build the model
    diresa = build_diresa(input_shape=(3,), dense_units=(40, 20, 2),
                          kernel_initializer="he_normal", kernel_regularizer=L2(1e-4),
                          dist_layer=DistLayer(dim_less=True))
    cov_weight = K.variable(0.)
    diresa.compile(loss=['MSE', LatentCovLoss(cov_weight), mse_dist_loss], loss_weights=[1., 1., 1.5])

    # train the model
    callback = [LossWeightAnnealing(cov_weight, target_loss=0.0001, anneal_step=0.2,
                                    start_epoch=3)]
    diresa.fit((train, train_twin), (train, train, train),
               validation_data=((val, val_twin), (val, val, val)),
               epochs=5, batch_size=512, shuffle=True, verbose=2, callbacks=callback)

    # decoder submodel
    cut_sub_model(diresa, 'Recon')

    # save and open model
    diresa.save('diresa.keras')
    load_model('diresa.keras')


if __name__ == "__main__":
    test_diresa()



#!/usr/bin/env python3
"""
Creates DIRESA and AE models out of an encoder and decoder model.
Creates DIRESA and AE models from hyperparameters.

:Author:  Geert De Paepe
:Email:   geert.de.paepe@vub.be
:License: MIT License

1. Creating AE and Diresa models out of an encoder and decoder model:
  - autoencoder_model(x, encoder, decoder)
  - diresa_model(x, x_twin, encoder, decoder)

2. Creating AE and Diresa models from hyperparameters
  - build_ae(input_shape, stack, stack_filters, latent_filters, kernel_size=(3, 3),
    conv_transpose=False, up_first=False, residual=False, dense_units=(), dropout_rate=0,
    activation='relu', encoder_activation='linear', decoder_activation='linear', **kwargs)
  - build_diresa(input_shape, stack, stack_filters, latent_filters, kernel_size=(3, 3),
    conv_transpose=False, up_first=False, residual=False, dense_units=(), dropout_rate=0,
    activation='relu', encoder_activation='linear', decoder_activation='linear', **kwargs)

   Encoder:
    - 0 or more [blocks] with C (Conv2D) or residual units and a P (MaxPooling layer)
    - 0 or 1 [block] of D (Dense layers)
   Decoder:
    - 0 or 1 [block] with D (Dense layers)
    - 0 or more [blocks] with C (Conv2D) or residual units and an U (UpSampling layer)
   Examples:
    - stack;     dense_units;  Encoder;                Decoder (up_first=True);    Decoder (up_first=False)
    - [1];       ();           [C-P]-Cout;             [U-C]-Cout;                 [C-U]-Cout
    - [3];       ();           [C-C-C-P]-Cout;         [U-C-C-C]-Cout;             [C-U-C-C]-Cout
    - [1,1];     ();           [C-P]-[C-P]-Cout;       [U-C]-[U-C]-Cout;           [C-U]-[C-U]-Cout
    - ();        [20,10];      [D-Dout];               [D-Dout];                   [D-Dout]
    - [2];       [20,10];      [C-C-P]-[D-Dout];       [D-D]-[U-C]-Cout;           [D-D]-[C-U]-Cout
    - [1,1];     [20,10];      [C-P]-[C-P]-[D-Dout];   [D-D]-[U]-[U-C]-Cout;       [D-D]-[U]-[C-U]-Cout

   If conv_transpose=True, C is a ConvTranspose layer, only possible for up_first=True.
   If residual=True, C is a ResNet V1 residual unit with a skip connection, only possible for up_first=True.
   If dropout_rate!=0, Dropout(dropout_rate) is added between Dense units, only possible if len(dense_units)>0.
   kwargs are passed to all Conv2D and Dense layers, e.g. they can be used for changing kernel_initializer or kernel_regularizer.
   Input rank should be 3 if Conv2D blocks, first 2 dimensions of input_shape should be a multiple of 2^len(stack).
   Input rank should be 1 if only a Dense block.
"""

from inspect import isfunction
from math import prod as _prod
from sys import exit
from tensorflow.keras import layers, Input
from tensorflow.keras.models import Model
from diresa.layers import DistLayer


def autoencoder_model(x, encoder, decoder):
    """
    Creates autoencoder model out of an encoder and a decoder model
    
    :param x: keras input tensor (keras.Input())
    :param encoder: encoder functional Keras model
    :param decoder: decoder functional Keras model
    :return: autoencoder model
    """
    y = encoder(x)
    y = decoder(y)
    return Model(x, y)


def diresa_model(x, x_twin, encoder, decoder, dist_layer=DistLayer()):
    """
    Creates a Diresa model out of an encoder and a decoder model
    
    :param x: keras input tensor (keras.Input())
    :param x_twin: keras input tensor for shuffled input
    :param encoder: encoder functional Keras model
    :param decoder: decoder functional Keras model
    :param dist_layer: distance layer to be used
    :return: Diresa model
    """
    latent_orig = encoder(x)
    latent_twin = encoder(x_twin)
    dist = dist_layer(x, x_twin, latent_orig, latent_twin)
    latent = layers.Flatten(name="Cov")(latent_orig)
    output = decoder(latent_orig)
    return Model((x, x_twin), (output, latent, dist))


def build_ae(input_shape=(),
             stack=(),
             stack_filters=(),
             latent_filters=1,
             kernel_size=(3, 3),
             conv_transpose=False,
             up_first=False,
             residual=False,
             dropout_rate=0,
             dense_units=(),
             activation='relu',
             activation_layer_param=None,
             encoder_activation='linear',
             decoder_activation='linear',
             **kwargs,
             ):
    """
    Creates an AE model out of hyperparameters
    
    :param input_shape: rank 3 with Conv2D layers, first 2 dims should be a multiple of 2^len(stack); rank 1 if only Dense layers
    :param stack: elements are nbr of Conv2D or residual units in a block
    :param stack_filters: elements are nbr of filters in a block
    :param latent_filters: nbr of filters in convolutional output (only used if no dense units)
    :param kernel_size: kernel size for convolution
    :param conv_transpose: if True ConvTranspose is used in decoder, only possible for up_first=True
    :param up_first: if True UpSampling is first in decoder block, if False UpSampling is second
    :param residual: if True, elements in blocks are residual units, if False elements are Conv2D layers
    :param dense_units: elements are nbr of nodes of a Dense layer in the dense block
    :param dropout_rate: if dropout_rate!=0, Dropout(dropout_rate) is added between Dense units, only if len(dense_units)>0
    :param activation: activation function or layer used (except for output of encoder/decoder)
    :param activation_layer_param: parameters for activation layer
    :param encoder_activation: activation function used for output of encoder
    :param decoder_activation: activation function used for output of decoder
    :param kwargs: are passed to all Conv2D and Dense layers, e.g. can be used for changing kernel_initializer, kernel_regularizer
    :return: AE functional Keras model
    """

    activation, activation_layer = _activation(activation)
    encoder, decoder = _encoder_decoder_model(input_shape=input_shape,
                                              stack=stack,
                                              stack_filters=stack_filters,
                                              latent_filters=latent_filters,
                                              kernel_size=kernel_size,
                                              conv_transpose=conv_transpose,
                                              up_first=up_first,
                                              residual=residual,
                                              dense_units=dense_units,
                                              dropout_rate=dropout_rate,
                                              activation=activation,
                                              activ_layer=activation_layer,
                                              activ_layer_param=activation_layer_param,
                                              encoder_activation=encoder_activation,
                                              decoder_activation=decoder_activation,
                                              **kwargs,
                                              )
    x = Input(shape=input_shape)
    return autoencoder_model(x, encoder, decoder)


def build_diresa(input_shape=(),
                 stack=(),
                 stack_filters=(),
                 latent_filters=1,
                 kernel_size=(3, 3),
                 conv_transpose=False,
                 up_first=False,
                 residual=False,
                 dropout_rate=0,
                 dense_units=(),
                 activation='relu',
                 activation_layer_param=None,
                 encoder_activation='linear',
                 decoder_activation='linear',
                 dist_layer=DistLayer(),
                 **kwargs,
                 ):
    """
    Creates a Diresa model out of hyperparameters
    
    :param input_shape: rank 3 with Conv2D layers, first 2 dims should be a multiple of 2^len(stack); rank 1 if only Dense layers
    :param stack: elements are nbr of Conv2D or residual units in a block
    :param stack_filters: elements are nbr of filters in a block
    :param latent_filters: nbr of filters in convolutional output (only used if no dense units)
    :param kernel_size: kernel size for convolution
    :param conv_transpose: if True ConvTranspose is used in decoder, only possible for up_first=True
    :param up_first: if True UpSampling is first in decoder block, if False UpSampling is second
    :param residual: if True, elements in blocks are residual units, if False elements are Conv2D layers
    :param dense_units: elements are nbr of nodes of a Dense layer in the dense block
    :param dropout_rate: if dropout_rate!=0, Dropout(dropout_rate) is added between Dense units, only if len(dense_units)>0
    :param activation: activation function or layer used (except for output of encoder/decoder)
    :param activation_layer_param: parameters for activation layer
    :param encoder_activation: activation function used for output of encoder
    :param decoder_activation: activation function used for output of decoder
    :param dist_layer: distance layer to be used
    :param kwargs: are passed to all Conv2D and Dense layers, e.g. can be used for changing kernel_initializer, kernel_regularizer
    :return: Diresa functional Keras model
    """

    activation, activation_layer = _activation(activation)
    encoder, decoder = _encoder_decoder_model(input_shape=input_shape,
                                              stack=stack,
                                              stack_filters=stack_filters,
                                              latent_filters=latent_filters,
                                              kernel_size=kernel_size,
                                              conv_transpose=conv_transpose,
                                              up_first=up_first,
                                              residual=residual,
                                              dense_units=dense_units,
                                              dropout_rate=dropout_rate,
                                              activation=activation,
                                              activ_layer=activation_layer,
                                              activ_layer_param=activation_layer_param,
                                              encoder_activation=encoder_activation,
                                              decoder_activation=decoder_activation,
                                              **kwargs,
                                              )
    input_orig = Input(name="Input", shape=input_shape)
    input_twin = Input(name="Shuffled_Input", shape=input_shape)
    return diresa_model(input_orig, input_twin, encoder, decoder, dist_layer=dist_layer)


#
# Helper functions for creating the encoder and decoder models out of the hyperparameters
#
def _activation(activation):
    if isfunction(activation) or isinstance(activation, str):
        activation_layer = None
    elif issubclass(activation, layers.Layer):
        activation_layer = activation
        activation = None
    else:
        print("Activation should be a function, string or layer")
        exit(1)
    return activation, activation_layer


def _residual_unit(x, filters, kernel_size, activation, activ_layer, activ_layer_param, name, **kwargs):
    y = layers.Conv2D(filters, kernel_size, padding='same', name=name + '_A', **kwargs)(x)
    y = layers.BatchNormalization()(y)
    if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
    if activation is not None: y = layers.Activation(activation)(y)
    y = layers.Conv2D(filters, kernel_size, padding='same', name=name + '_B', **kwargs)(y)
    y = layers.BatchNormalization()(y)

    # If input has a  different nbr of filters
    if x.shape[-1] != filters:
        x = layers.Conv2D(filters, (1, 1), padding='same', name=name + '_C', **kwargs)(x)

    # Add skip connection
    y = layers.Add(name=name.replace("Conv2D", "Add"))([x, y])

    return layers.Activation(activation)(y)


def _encoder_block(y, block, filters, kernel_size, residual, activation, activ_layer, activ_layer_param, name, **kwargs):
    # Block of num residual or Conv2D units followed by a MaxPooling2D layer
    for unit in range(block):
        if residual:
            y = _residual_unit(y, filters, kernel_size, activation=activation, activ_layer=activ_layer,
                               activ_layer_param=activ_layer_param, name=name + "_Conv2D_" + str(unit), **kwargs)
        else:
            y = layers.Conv2D(filters, kernel_size, activation=activation, padding='same',
                              name=name + "_Conv2D_" + str(unit), **kwargs)(y)
            if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
    y = layers.MaxPooling2D((2, 2), padding='same', name=name + '_MaxPooling2D')(y)
    return y


def _decoder_block(y, block, filters, kernel_size, conv_transpose, residual, activation, activ_layer, activ_layer_param, up_first,
                   name, **kwargs):
    if up_first:
        # Block of an UpSampling2D layer followed by num residual or Conv2D units
        y = layers.UpSampling2D((2, 2), name=name + '_UpSampling2D')(y)
        for unit in range(block):
            if residual:
                y = _residual_unit(y, filters, kernel_size, activation=activation, activ_layer=activ_layer,
                                   activ_layer_param=activ_layer_param, name=name + "_Conv2D_" + str(unit), **kwargs)
            elif conv_transpose:
                y = layers.Conv2DTranspose(filters, kernel_size, activation=activation, padding='same',
                                           name=name + "_Conv2DTranspose_" + str(unit), **kwargs)(y)
                if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
            else:
                y = layers.Conv2D(filters, kernel_size, activation=activation, padding='same',
                                  name=name + "_Conv2D_" + str(unit), **kwargs)(y)
                if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
    else:
        # Block of a Conv2D layer followed by an UpSampling2D layer followed by num-1 Conv2D layers
        if block > 0:
            y = layers.Conv2D(filters, kernel_size, activation=activation, padding='same',
                              name=name + '_Conv2D_0', **kwargs)(y)
            if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
        y = layers.UpSampling2D((2, 2), name=name + '_UpSampling2D')(y)
        for unit in range(1, block):
            y = layers.Conv2D(filters, kernel_size, activation=activation, padding='same',
                              name=name + '_Conv2D_' + str(unit), **kwargs)(y)
            if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
    return y


def _encoder_model(input_shape, stack, stack_filters, latent_filters, kernel_size, conv_transpose, up_first, residual,
                   dense_units, dropout_rate, activation, activ_layer, activ_layer_param, encoder_activation, **kwargs):
    if len(stack) == 0 and len(dense_units) == 0:
        print("You should have minimum 1 convolutional or 1 dense layer")
        exit(1)
    if len(stack) > 1 and len(input_shape) != 3:
        print("Length input_shape should be 3 with convolutional layers")
        exit(1)
    if len(stack) == 0 and len(input_shape) != 1:
        print("Length input_shape should be 1 if only dense layers")
        exit(1)
    if len(stack) != len(stack_filters):
        print("stack and stack_filters should have the same length")
        exit(1)
    if len(stack) > 1 and input_shape[0] % (2 ** len(stack)) != 0:
        print("input_shape[0] should be a multiple of 2^len(stack)")
        exit(1)
    if len(stack) > 1 and input_shape[1] % (2 ** len(stack)) != 0:
        print("input_shape[1] should be a multiple of 2^len(stack)")
        exit(1)
    if not up_first and residual:
        print("Residual only possible with UpSampling layer first in decoder")
        exit(1)
    if residual and conv_transpose:
        print("Residual not possible with Conv2DTranspose layer in decoder")
        exit(1)
    if not up_first and conv_transpose:
        print("Transposed convolution only possible with UpSampling layer first in decoder")
        exit(1)
    if dropout_rate != 0 and len(dense_units) < 2:
        print("Dropout only possible if more than 1 dense layer")
        exit(1)

    x = Input(input_shape, name="Encoder_Input")
    y = x

    # Encoder blocks with Conv2D or residual units and a MaxPooling layer
    block_nr = 1
    for block, filters in zip(stack, stack_filters):
        y = _encoder_block(y, block=block, filters=filters, kernel_size=kernel_size, residual=residual, activation=activation,
                           activ_layer=activ_layer, activ_layer_param=activ_layer_param, name='Enc_' + str(block_nr), **kwargs)
        block_nr += 1

    # Encoder dense layers
    if len(dense_units) != 0:
        if len(stack) > 0:
            y = layers.Flatten()(y)
        for layer, units in enumerate(dense_units):
            if layer != len(dense_units) - 1:
                y = layers.Dense(units, activation=activation,
                                 name='Enc_' + str(block_nr) + '_Dense_' + str(layer), **kwargs)(y)
                if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
                if dropout_rate != 0:
                    y = layers.Dropout(dropout_rate)(y)
            else:  # last layer has other activation
                y = layers.Dense(units, activation=encoder_activation, name='Dense_Latent', **kwargs)(y)
    # If no dense layers, last Conv2D layer
    else:
        y = layers.Conv2D(latent_filters, kernel_size, activation=encoder_activation, padding='same',
                          name='Enc_' + str(block_nr) + '_Conv2D_0', **kwargs)(y)
        y = layers.Flatten(name='Flatten_Latent')(y)

    model = Model(x, y, name="Encoder")
    return model


def _decoder_model(input_shape, stack, stack_filters, latent_filters, kernel_size, conv_transpose, up_first, residual,
                   dense_units, dropout_rate, activation, activ_layer, activ_layer_param, decoder_activation, **kwargs):
    # Nbr of blocks in en/decoder
    block_nr = len(stack) + 1
    # Input shape of decoder
    comp_factor = 2 ** len(stack_filters)
    if len(dense_units) == 0 and len(stack) != 0:  # only convolutional layers
        conv_output_shape = (input_shape[0] // comp_factor, input_shape[1] // comp_factor, latent_filters)
        decoder_input_shape = (_prod(conv_output_shape),)
    elif len(stack) != 0:  # convolutional and dense layers
        decoder_input_shape = (dense_units[-1],)
        conv_output_shape = (input_shape[0] // comp_factor, input_shape[1] // comp_factor, stack_filters[-1])
    else:  # only dense layers
        decoder_input_shape = (dense_units[-1],)

    x = Input(shape=decoder_input_shape, name="Decoder_Input")
    y = x

    # Decoder dense layers
    if len(dense_units) != 0:
        if len(dense_units) > 1:
            for layer, units in enumerate(dense_units[-2::-1]):
                y = layers.Dense(units, activation=activation,
                                 name='Dec_' + str(block_nr) + '_Dense_' + str(len(dense_units) - layer - 1), **kwargs)(y)
                if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
                if dropout_rate != 0:
                    y = layers.Dropout(dropout_rate)(y)
        if len(stack) != 0:
            # Last dense layer units should match conv layer
            y = layers.Dense(_prod(conv_output_shape), activation=activation,
                             name='Dec_' + str(block_nr) + '_Dense_0', **kwargs)(y)
            if activ_layer is not None: y = activ_layer(**activ_layer_param)(y)
            # Shape of encoder output after convolution
            y = layers.Reshape(conv_output_shape)(y)
        else:
            # Last dense layer if no conv layers
            y = layers.Dense(input_shape[0], activation=decoder_activation, name='Dec_' + str(block_nr) + '_Dense_0', **kwargs)(y)
        block_nr -= 1
    else:
        y = layers.Reshape(conv_output_shape)(y)

    # Decoder blocks with Conv2D layers or residual elements and an UpSampling layer
    for block, filters in zip(stack[::-1], stack_filters[::-1]):
        if len(dense_units) != 0 and block_nr == len(stack):
            # Fist block has 1 conv layer less in case of a dense block
            y = _decoder_block(y, block=block - 1, filters=filters, conv_transpose=conv_transpose, kernel_size=kernel_size,
                               residual=residual, activation=activation, activ_layer=activ_layer,
                               activ_layer_param=activ_layer_param, up_first=up_first, name='Dec_' + str(block_nr), **kwargs)
        else:
            y = _decoder_block(y, block=block, filters=filters, conv_transpose=conv_transpose, kernel_size=kernel_size,
                               residual=residual, activation=activation, activ_layer=activ_layer,
                               activ_layer_param=activ_layer_param, up_first=up_first, name='Dec_' + str(block_nr), **kwargs)
        block_nr -= 1

    # Last Conv2D layer
    if len(stack) != 0:
        y = layers.Conv2D(input_shape[-1], kernel_size, activation=decoder_activation, padding='same',
                          name='Dec_' + str(block_nr) + '_Conv2D_0', **kwargs)(y)

    model = Model(x, y, name="Recon")
    return model


def _encoder_decoder_model(input_shape, stack, stack_filters, latent_filters, kernel_size, conv_transpose, up_first,
                           residual, dense_units, dropout_rate, activation, activ_layer, activ_layer_param, encoder_activation,
                           decoder_activation, **kwargs):
    encoder = _encoder_model(input_shape=input_shape, stack=stack, stack_filters=stack_filters, latent_filters=latent_filters,
                             kernel_size=kernel_size, conv_transpose=conv_transpose, up_first=up_first, residual=residual,
                             dense_units=dense_units, dropout_rate=dropout_rate, activation=activation, activ_layer=activ_layer,
                             activ_layer_param=activ_layer_param, encoder_activation=encoder_activation, **kwargs)
    decoder = _decoder_model(input_shape=input_shape, stack=stack, stack_filters=stack_filters, latent_filters=latent_filters,
                             kernel_size=kernel_size, conv_transpose=conv_transpose, up_first=up_first, residual=residual,
                             dense_units=dense_units, dropout_rate=dropout_rate, activation=activation, activ_layer=activ_layer,
                             activ_layer_param=activ_layer_param, decoder_activation=decoder_activation, **kwargs)
    return encoder, decoder



#!/usr/bin/env python3
"""
DIRESA loss classes/functions

:Author:  Geert De Paepe
:Email:   geert.de.paepe@vub.be
:License: MIT License
"""

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.losses import Loss
from diresa.toolbox import covariance


@tf.keras.utils.register_keras_serializable()
class LatentCovLoss(Loss):
    """
    Latent covariance loss class
    Latent covariance weight is annealed by AnnealingCallback
    """
    def __init__(self, cov_weight=K.variable(1.), **kwargs):
        """
        :param cov_weight: tensorflow variable with initial covariance loss weight
        """
        super().__init__(**kwargs)
        self.cov_weight = cov_weight

    def call(self, _, latent):
        """
        :param _: not used (loss functions need 2 params: the true and predicted values)
        :param latent: batch of latent vectors
        :return: weighted covariance loss
        """
        cov = tf.math.abs(covariance(latent))
        cov_square = tf.math.multiply(cov, cov)
        nbr_of_cov = tf.shape(latent)[-1] * (tf.shape(latent)[-1] - 1)
        cov_loss = (tf.math.reduce_sum(cov_square) - tf.linalg.trace(cov_square)) / tf.cast(nbr_of_cov, "float32")
        return self.cov_weight * cov_loss

    def get_config(self):
        config = super().get_config()
        config.update({'cov_weight': self.cov_weight.numpy()})
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


@tf.keras.utils.register_keras_serializable()
def mae_dist_loss(_, distances):
    """
    Absolute Error between original and latent distances
    
    :param _: not used (loss functions need 2 params: the true and predicted values)
    :param distances: batch of original and latent distances between twins
    :return: batch of absolute errors
    """
    ae = tf.math.abs(distances[:, 0] - distances[:, 1])
    return ae


@tf.keras.utils.register_keras_serializable()
def male_dist_loss(_, distances):
    """
    Absolute Error between logarithm of original and latent distances
    
    :param _: not used (loss functions need 2 params: the true and predicted values)
    :param distances: batch of original and latent distances between twins
    :return: batch of absolute logarithmic errors
    """
    ale = tf.math.abs(tf.math.log1p(distances[:, 0]) - tf.math.log1p(distances[:, 1]))
    return ale


@tf.keras.utils.register_keras_serializable()
def mape_dist_loss(_, distances):
    """
    Absolute Percentage Error between original and latent distances
    
    :param _: not used (loss functions need 2 params: the true and predicted values)
    :param distances: batch of original and latent distances between twins
    :return: batch of absolute percentage errors
    """
    epsilon = 1e-8
    ape = tf.math.abs((distances[:, 0] - distances[:, 1]) / (distances[:, 0] + epsilon))
    return ape


@tf.keras.utils.register_keras_serializable()
def mse_dist_loss(_, distances):
    """
    Squared Error between original and latent distances
    
    :param _: not used (loss functions need 2 params: the true and predicted values)
    :param distances: batch of original and latent distances between twins
    :return: batch of squared errors
    """
    se = tf.math.square(distances[:, 0] - distances[:, 1])
    return se


@tf.keras.utils.register_keras_serializable()
def msle_dist_loss(_, distances):
    """
    Squared Error between logarithm of original and latent distances
    
    :param _: not used (loss functions need 2 params: the true and predicted values)
    :param distances: batch of original and latent distances between twins
    :return: batch of squared logarithmic errors
    """
    sle = tf.math.square(tf.math.log1p(distances[:, 0]) - tf.math.log1p(distances[:, 1]))
    return sle


@tf.keras.utils.register_keras_serializable()
def corr_dist_loss(_, distances):
    """
    Correlation loss between original and latent distances
    
    :param _: not used (loss functions need 2 params: the true and predicted values)
    :param distances: batch of original and latent distances between twins
    :return: 1 - correlation coefficient
    """
    cov = covariance(distances)
    cov_sqrt = tf.math.sqrt(tf.math.abs(cov))
    return 1 - cov[0, 1] / (cov_sqrt[0, 0] * cov_sqrt[1, 1])


@tf.keras.utils.register_keras_serializable()
def corr_log_dist_loss(_, distances):
    """
    Correlation loss between logarithm of original and latent distances
    
    :param _: not used (loss functions need 2 params: the true and predicted values)
    :param distances: batch of original and latent distances between twins
    :return: 1 - correlation coefficient (of logarithmic distances)
    """
    cov = covariance(tf.math.log1p(distances))
    cov_sqrt = tf.math.sqrt(tf.math.abs(cov))
    return 1 - cov[0, 1] / (cov_sqrt[0, 0] * cov_sqrt[1, 1])



#!/usr/bin/env python3
"""
DIRESA custom layer classes
:Author:  Geert De Paepe
:Email:   geert.de.paepe@vub.be
:License: MIT License
"""

import tensorflow as tf
from tensorflow.keras import layers


@tf.keras.utils.register_keras_serializable()
class DistLayer(layers.Layer):
    """
    Calculates distances between the 2 inputs and distances between the 2 latent representations of a twin model
    """

    def __init__(self, dim_less=False, name="Dist", **kwargs):
        """
        :param dim_less: if True distance is divided by dimension of space
        """
        super(DistLayer, self).__init__(name=name, **kwargs)
        self.dim_less = dim_less

    def call(self, x1, x2, y1, y2):
        """
        :param x1: batch of input samples to encoder
        :param x2: batch of input shuffled samples to twin encoder
        :param y1: batch of latent representations of encoder
        :param y2: batch of latent representations of twin encoder
        :return: batch of distances between inputs, batch of distances between latent representations
        """
        dist1 = tf.math.square(x1 - x2)
        dist2 = tf.math.square(y1 - y2)
        dist1 = tf.reduce_sum(tf.reshape(dist1, [tf.shape(dist1)[0], -1]), axis=1)  # sum over all dims, except 0
        dist2 = tf.reduce_sum(tf.reshape(dist2, [tf.shape(dist2)[0], -1]), axis=1)  # sum over all dims, except 0
        if self.dim_less:
            dim1 = tf.cast(tf.math.divide(tf.size(x1), tf.shape(x1)[0]), dtype=tf.float32)
            dist1 = tf.math.divide(dist1, dim1)  # divide by input dimension
            dim2 = tf.cast(tf.math.divide(tf.size(y1), tf.shape(y1)[0]), dtype=tf.float32)
            dist2 = tf.math.divide(dist2, dim2)  # divide by latent space dimension
        return tf.stack((dist1, dist2), axis=1)

    def get_config(self):
        config = super().get_config()
        config.update({'dim_less': self.dim_less})
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


@tf.keras.utils.register_keras_serializable()
class OrderLayer(layers.Layer):
    """
    Reorders the inputs according to a given order
    """

    def __init__(self, order, name=None, **kwargs):
        super(OrderLayer, self).__init__(name=name, **kwargs)
        self.order = order

    def call(self, x):
        """
        :param x: batch of inputs
        :return: batch with reordered inputs
        """
        y = [x[:, i] for i in self.order]
        return y

    def get_config(self):
        config = super().get_config()
        config.update({'order': self.order})
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



#!/usr/bin/env python3
"""
DIRESA helper functions

:Author:  Geert De Paepe
:Email:   geert.de.paepe@vub.be
:License: MIT License
"""

from sys import exit
import tensorflow as tf
from tensorflow.keras.models import Model
from diresa.layers import OrderLayer


def covariance(x):
    """
    Computes the covariance matrix of x (normalisation is divided by N)
    :param x: 2-D array, row are variables, columns are samples
    :return: covariance matrix
    """
    mean_x = tf.expand_dims(tf.reduce_mean(x, axis=0), 0)
    mx = tf.matmul(tf.transpose(mean_x), mean_x)
    vx = tf.matmul(tf.transpose(x), x)/tf.cast(tf.shape(x)[0], tf.float32)
    cov_xx = vx - mx
    return cov_xx


def r2_score(y, y_pred):
    """
    :param y: original dataset
    :param y_pred: predicted dataset
    :return: R2 score between y and y_pred
    """
    error = tf.math.reduce_sum(tf.math.square(y - y_pred))
    var = tf.math.reduce_sum(tf.math.square(y - tf.math.reduce_mean(y, axis=0)))
    r2 = 1.0 - error / var
    return r2


def set_components_to_mean(latent, untouched=(0,)):
    """
    Sets all latent components to mean except the ones in the list (which are kept untouched)
    Limitations: assumes a flat latent space
    :param latent: latent dataset
    :param untouched: components not in this list are set to mean
    :return: latent dataset with all components set to mean except the ones in the list
    """
    for i in range(0, latent.shape[1]):
        if i not in untouched:
            latent[:, i] = tf.math.reduce_mean(latent[:, i])
    return latent


def cut_sub_model(model, sub_model_name):
    """
    Cuts a sub-model out of a keras model 
    Limitations: does not work for a sub-model of a sub-model
    :param model: keras model
    :param sub_model_name: name of the sub-model
    :return: submodel
    """
    sub_model_nbr = None
    sub_model_config = None

    for nbr, layer in enumerate(model.get_config()['layers']):
        if layer['name'] == sub_model_name:
            sub_model_config = layer['config']
            sub_model_nbr = nbr

    if sub_model_config is None:
        print(sub_model_name, " not found in model")
        exit(1)

    sub_model = Model.from_config(sub_model_config)
    weights = [layer.get_weights() for layer in model.layers[sub_model_nbr].layers[1:]]

    for layer, weight in zip(sub_model.layers[1:], weights):
        layer.set_weights(weight)

    return sub_model


def order_latent(encoder, decoder, dataset):
    latent = encoder.predict(dataset, verbose=0)
    latent_dim = latent.shape[1]

    score = []
    for component in range(latent_dim):
        latent_component = set_components_to_mean(latent, untouched=(component,))
        decoded_component = decoder.predict(latent_component, verbose=0)
        score.append(r2_score(dataset, decoded_component))

    ranking = sorted(range(len(score)), key=score.__getitem__)
    reverse_ranking = sorted(range(len(ranking)), key=ranking.__getitem__)
    ordered_score = [score[i] for i in ranking]
    return ranking, reverse_ranking, ordered_score


def encoder_decoder(model, dataset=None, encoder_name="Encoder", decoder_name="Decoder"):
    """
    Returns encoder and decoder out of DIRESA model
    If dataset is not None: adds ordering layers after encoder and before decoder
    :param model: keras model
    :param dataset: dataset
    :param encoder_name: name of the encoder
    :param decoder_name: name of the decoder
    :return: encoder and decoder model
    """
    encoder = cut_sub_model(model, encoder_name)
    decoder = cut_sub_model(model, decoder_name)

    if dataset is not None:
        order, reverse_order, _ = order_latent(encoder, decoder, dataset)
        order_layer = OrderLayer(order)
        reverse_order_layer = OrderLayer(reverse_order)
        encoder = Model(inputs=encoder.inputs, outputs=order_layer(encoder.outputs))
        decoder = Model(inputs=reverse_order_layer.inputs, outputs=decoder(reverse_order_layer.outputs))

    return encoder, decoder



#!/usr/bin/env python3
"""
:Author:  Geert De Paepe
:Email:   geert.de.paepe@vub.be
:License: MIT License
"""

from diresa.models import build_diresa
from tensorflow.keras.activations import relu
from tensorflow.keras.layers import PReLU
from tensorflow.keras.initializers import Constant
from diresa.toolbox import cut_sub_model


def test_build_models():
    hyper_params = [{"input_shape": [8, 8, 1], "stack": [1, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": False},
                    {"input_shape": [8, 8, 1], "stack": [1, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": True},
                    {"input_shape": [8, 8, 1], "stack": [1, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": True, "conv_transpose": True, "activation": "sigmoid"},
                    {"input_shape": [8, 8, 1], "stack": [1, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": True, "residual": True},
                    {"input_shape": [8, 8, 1], "stack": [3, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": False},
                    {"input_shape": [8, 8, 1], "stack": [3, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": True, "activation": relu},
                    {"input_shape": [8, 8, 1], "stack": [3, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": True, "conv_transpose": True},
                    {"input_shape": [8, 8, 1], "stack": [3, ], "stack_filters": [32, ],
                     "dense_units": (), "up_first": True, "residual": True},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": (), "up_first": False,
                     "activation": PReLU, "activation_layer_param": {"alpha_initializer": Constant(-1.), "shared_axes": [1, 2]}},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": (), "up_first": True},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": (), "up_first": True, "conv_transpose": True},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": (), "up_first": True, "residual": True},
                    {"input_shape": [20, ], "dense_units": [15, 10, 5]},
                    {"input_shape": [20, ], "stack": [], "stack_filters": [], "dense_units": (15, 10, 5),
                     "activation": PReLU, "activation_layer_param": {"alpha_initializer": Constant(-1.), "shared_axes": [1]}},
                    {"input_shape": [8, 8, 1], "stack": [2, ], "stack_filters": [32, ],
                     "dense_units": [20, 10], "up_first": False},
                    {"input_shape": [8, 8, 1], "stack": [2, ], "stack_filters": [32, ],
                     "dense_units": [20, 10], "up_first": True},
                    {"input_shape": [8, 8, 1], "stack": [2, ], "stack_filters": [32, ],
                     "dense_units": [20, 10], "up_first": True, "conv_transpose": True},
                    {"input_shape": [8, 8, 1], "stack": [2, ], "stack_filters": [32, ],
                     "dense_units": [20, 10], "up_first": True, "residual": True,
                     "activation": PReLU, "activation_layer_param": {"alpha_initializer": Constant(-1.)}},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": [20, 10], "up_first": False},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": [20, 10], "up_first": True},
                    {"input_shape": [8, 8, 1], "stack": [1, 2], "stack_filters": [32, 16],
                     "dense_units": [20, 10], "up_first": True, "conv_transpose": True},
                    {"input_shape": [8, 8, 1], "stack": [1, 2], "stack_filters": [32, 16],
                     "dense_units": [20, 10], "up_first": True, "residual": True},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": [20, 10, 10], "dropout_rate": 0.25, "activation": relu},
                    {"input_shape": [8, 8, 1], "stack": [1, 1], "stack_filters": [32, 16],
                     "dense_units": [20, 10, 10], "kernel_initializer": "he_normal", "kernel_regularizer": "L2"},
                    ]
    for hyper_param in hyper_params:
        print("\n\n", hyper_param)
        diresa = build_diresa(**hyper_param)
        diresa.summary(expand_nested=True)


if __name__ == "__main__":
    test_build_models()
